'use strict';

/// <reference path="../../assets/jquery.d.ts" />

//Note that we use the WordPress version of Lodash here, not the one from the plugin.
declare const lodash;
declare const wsAdminCustomizerPreview;

declare const wsAmeBrandingClsPreviewData: {
	/**
	 * Setting ID -> Color variable name
	 */
	colorSettings: {
		[settingId: string]: {
			initialValue: string | null;
			variableName: string;
		}
	};
	/**
	 * Color variable name -> index
	 */
	colorVariableOrder: Record<string, number>;
	previewBaseUrl: string;

	isOverrideEnabled: boolean;
	overrideSettingId: string;
	forceSettingId: string;
	forceEnablePreview: boolean;
};

(function ($) {
	if (typeof lodash === 'undefined') {
		return;
	}

	let isOverrideEnabled = wsAmeBrandingClsPreviewData.isOverrideEnabled;
	let forceEnablePreview = wsAmeBrandingClsPreviewData.forceEnablePreview;

	let currentColors: { [variableName: string]: string | null } = {};

	let $nodesToRemove: JQuery[] = [];
	let $latestStylesheet: JQuery = null;

	const updateColorPreview = lodash.throttle(function () {
		//Preview is enabled only if the "is_color_override_enabled" setting is enabled
		//or if the special "force enable preview" setting is enabled.
		const isPreviewEnabled = isOverrideEnabled || forceEnablePreview;
		if (!isPreviewEnabled) {
			//Remove all preview stylesheets.
			$nodesToRemove.forEach(function ($node) {
				$node.remove();
			});
			$nodesToRemove = [];
			if ($latestStylesheet) {
				$latestStylesheet.remove();
				$latestStylesheet = null;
			}

			//Also remove the normal color scheme stylesheet generated by PHP
			//if it looks like it was added by this plugin.
			const $colorSchemeStyle = $('head link#colors-css');
			if ($colorSchemeStyle.length > 0) {
				const href = $colorSchemeStyle.attr('href');
				if (href && (href.indexOf('ame_branding_preview_colors')) !== -1) {
					$colorSchemeStyle.remove();
				}
			}

			return;
		}

		let colors = [];
		for (let variableName in currentColors) {
			let color = currentColors[variableName];
			if ((color !== null) && (color !== '')) {
				const index = wsAmeBrandingClsPreviewData.colorVariableOrder[variableName];
				colors[index] = color.replace('#', '');
			}
		}

		const stylesheetUrl = wsAmeBrandingClsPreviewData.previewBaseUrl
			+ '&colors=' + encodeURIComponent(colors.join('.'));

		const $stylesheet = $(
			'<link/>',
			{
				'rel': 'stylesheet',
				'type': 'text/css',
				'href': stylesheetUrl
			}
		);

		if ($latestStylesheet !== null) {
			$nodesToRemove.push($latestStylesheet);
		}
		$latestStylesheet = $stylesheet;

		$stylesheet.on('load', function () {
			if ($nodesToRemove.length > 0) {
				$nodesToRemove.forEach(function ($node) {
					$node.remove();
				});
				$nodesToRemove = [];
			}
		});

		$stylesheet.appendTo('head');
	}, 1000);

	//Initialize the current colors.
	for (const settingId in wsAmeBrandingClsPreviewData.colorSettings) {
		const setting = wsAmeBrandingClsPreviewData.colorSettings[settingId];
		currentColors[setting.variableName] = setting.initialValue;

		wsAdminCustomizerPreview.registerPreviewer(settingId, function (newColor: string | null) {
			currentColors[setting.variableName] = newColor;
			updateColorPreview();
		});
	}

	//Watch for changes to the "color scheme override" and "force enable preview" settings.
	if (wsAmeBrandingClsPreviewData.overrideSettingId) {
		wsAdminCustomizerPreview.registerPreviewer(
			wsAmeBrandingClsPreviewData.overrideSettingId,
			function (newValue: boolean) {
				isOverrideEnabled = !!newValue;
				updateColorPreview();
			}
		);
	}
	wsAdminCustomizerPreview.registerPreviewer(
		wsAmeBrandingClsPreviewData.forceSettingId,
		function (newValue: boolean) {
			forceEnablePreview = !!newValue;
			updateColorPreview();
		}
	);

})(jQuery);
